class Scale:
    # SCALE_FACTOR = 1 units are in mm
    # SCALE_FACTOR = 10 units are in 10 mm
    # SCALE_FACTOR = 100 units are in .1 m
    SCALE_FACTOR = 1
    LENGTH = SCALE_FACTOR
    AREA = SCALE_FACTOR * SCALE_FACTOR
    VOLUME = SCALE_FACTOR * SCALE_FACTOR * SCALE_FACTOR



from . import ArtefactUtilsParseException

from enum import Enum


class OrientationMethod(Enum):
    ORIENT_TIP_TO_BUTT = ("Tip2Butt", "T2B", "Tip to Butt")
    ORIENT_TIP_TO_COM = ("Tip2CoM", "T2C", "Tip to Centre of Mass")
    ORIENT_TIP_TO_COHM = ("Tip2CoHM", "T2H", "Tip to Centre of Hollow Mesh")
    ORIENT_TO_FITLINE = ("FitLine", "FIT", "Fit line through points")

    def __init__(
        self,
        text,
        abbreviation,
        description,
    ):
        self.text = text
        self.abbreviation = abbreviation
        self.description = description

    @staticmethod
    def from_text(text):
        for orientation_method in OrientationMethod:
            if text == orientation_method.text:
                return orientation_method
        # No matching orientation method found
        raise (
            ArtefactUtilsParseException(
                f"Unable to create OrintationMethod from text: {text}."
            )
        )



import numpy as np

PointXyzType = np.dtype((np.float64, 3))

AngleType = np.dtype(
    [
        ("full_angle", np.float64),
        ("quadrant_angle", np.float64),
        ("start_angle", np.int16),
        ("clockwise", np.bool_),
        ("breadth_sign", np.int8),
        ("thickness_sign", np.int8),
    ]
)

PointBltType = np.dtype(
    [
        ("breadth", np.float64),
        ("length", np.float64),
        ("thickness", np.float64),
    ]
)

SampleType = np.dtype(
    [
        ("angle", AngleType),
        ("point", PointBltType),
    ]
)



# Some links
# https://towardsdatascience.com/3d-data-processing-with-open3d-c3062aadc72e
# https://sketchfab.com/3Djimmetry/models visual reference for all artefacts

import csv
import logging.config
import pathlib
import sys
from typing import Iterator
import yaml

import filelock
import vedo


def range_include_stop(
    start: int,
    stop: int,
    step: int = 1,
) -> Iterator[int]:
    """Alternative to the python range() function which includes outputting the stop value"""
    # The python range function would normally exclued the
    # stop value. This function allows us to include the
    # stop value without adjusting the stop value to stop + step
    for i in range(start, stop, step):
        yield i
    # include stop if it would have been one of the steps
    if ((start - stop) % step) == 0:
        yield stop


def get_artefact_alignment_planes(
    artefact: vedo.Mesh,
) -> tuple[vedo.Plane, vedo.Plane, vedo.Plane]:
    """Make planes for all xy, xz, yz for viusalisations"""
    (
        breadth_min,
        breadth_max,
        length_min,
        length_max,
        thickness_min,
        thickness_max,
    ) = artefact.bounds()

    breadth = abs(breadth_max - breadth_min)
    length = abs(length_max - length_min)
    thickness = abs(thickness_max - thickness_min)

    breadth_pos = breadth_min + breadth / 2
    length_pos = length_min + length / 2
    thickness_pos = thickness_min + thickness / 2

    x_plane = vedo.Plane(
        pos=(0, length_pos, thickness_pos),
        normal=(1, 0, 0),
        s=(thickness, length),
        c="yellow",
        alpha=0.5,
    )

    z_plane = vedo.Plane(
        pos=(breadth_pos, length_pos, 0),
        normal=(0, 0, 1),
        s=(breadth, length),
        c="green",
        alpha=0.5,
    )

    y_plane = vedo.Plane(
        pos=(breadth_pos, 0, thickness_pos),
        normal=(0, 1, 0),
        s=(thickness, breadth),
        c="blue",
        alpha=0.5,
    )

    return x_plane, y_plane, z_plane


def output_csv_data(
    filepath: pathlib.Path,
    data_dict: dict,
):
    """Writes a dictionary to a file"""
    lock_file = filepath.with_suffix(filepath.suffix + ".lock")

    lock = filelock.FileLock(lock_file=lock_file)
    with lock:
        if not filepath.exists():
            out = open(filepath, "w", newline="", encoding="utf8")
            writer = csv.DictWriter(
                out, fieldnames=data_dict.keys(), quoting=csv.QUOTE_ALL
            )
            writer.writeheader()
        else:
            out = open(filepath, "a", newline="", encoding="utf8")
            writer = csv.DictWriter(
                out, fieldnames=data_dict.keys(), quoting=csv.QUOTE_ALL
            )

        writer.writerow(data_dict)
        out.close()

        if True:  # output the csv data to stdout if set to True
            stdout_writer = csv.DictWriter(
                sys.stdout, fieldnames=data_dict.keys(), quoting=csv.QUOTE_ALL
            )
            stdout_writer.writeheader()
            stdout_writer.writerow(data_dict)


def get_element_path(
    node: dict,
    prefix: pathlib.Path = None,
) -> Iterator[pathlib.Path]:
    """Reads a yaml file and builds a path from a heirarchical structure"""
    match node:
        case dict():
            for element in node:
                if prefix:
                    newprefix = pathlib.Path(prefix) / element
                else:
                    newprefix = pathlib.Path(element)
                yield from get_element_path(node[element], prefix=newprefix)
        case list():
            for element in node:
                yield from get_element_path(element, prefix=prefix)
        case str():
            output = pathlib.Path(prefix) / node
            yield (output)
        case _:
            pass  # excludes entries where the files are commented out


def setup_logging(
    config_file: pathlib.Path,
):
    """Configure logging using a config file"""

    with open(config_file, encoding="utf8") as f_in:
        config = yaml.safe_load(f_in)

    log_file = pathlib.Path(config["handlers"]["file"]["filename"])

    if not log_file.parent.exists():
        log_file.parent.mkdir(parents=True)

    logging.config.dictConfig(config)


if __name__ == "__main__":
    # TODO: add command line parsing to allow artefact processing from the command line

    exit()



"""Implementation of AnalysisMethod"""

from enum import Enum

from . import ArtefactUtilsParseException


class AnalysisMethod(Enum):
    """Enum defining the anaysis methods supported with associated meta data"""

    ENDO_SYM = ("Endo-Sym", "endo-sym", "magenta")
    MEAN_SYM = ("Mean-Sym", "mean-sym", "red")
    EXO_SYM = ("Exo-Sym", "exo-sym", "blue")

    def __init__(self, text, abbreviation, colour):
        self.text = text
        self.abbreviation = abbreviation
        self.colour = colour

    @staticmethod
    def from_text(text: str):
        """Method to set AnalysisMethod based on text matching associated meta data"""
        for analysis_method in AnalysisMethod:
            if text == analysis_method.text:
                return analysis_method
        # No matching orientation method found
        raise (
            ArtefactUtilsParseException(
                f"Unable to create AnalysisMethod from text: {text}."
            )
        )



"""
Evaluating Handaxe Bilateral/Bifacial Mirror Symmetry Along the Primary Axis in
3D

Any solid bifurcated by a plane along its primary axis can be split into two
shapes (in this case left and right halves or ventral and dorsal faces) with
equivalent volumes, but both halves may take an infinite number of shapes while
sharing the primary plane of reflection (the long axes), thus nullifying the
claim that correlated volumes of bifurcated stone tools can be used as a valid
warrant in the analoyses of mirror symmetry. In order for a particular volume to
act as a valid warrant for mirror symmetry, it must reference a known
symmetrical volume. In order to evaluate bilateral and bifacial mirror
symmetries in reference to the long axis (i.e., the Y-axis), this analyzer
generates a known symmetrical mesh within a shared analysis space oriented to
the same axes as the stone tool. The analysis assumes that a bifacially and
bilaterally symmetrical mesh generated to fit within (the Endo-Sym) or encompass
(the Exo-Sym) the artefact provides the necessary symmetrical reference to
warrant a volume-based analysis of mirror (reflection) symmetry, under the
assumption that within the same analysis space and oriented to the same axes,
any volume occupying the space beyond the surface of the Endo-Sym and beneath
the surface of the Exo-Sym is due to asymmetry inherent in the artefact.

Though the analyses are performed with the artefact in plan view (tip up, butt
down), to better visualize an evaluation of bilateral and bifacial mirror
symmetry, the tip of the artefact is rotated around the X-axis by -90 degrees in
the Z-plane to provide a bottom view of the artefact, with the butt centered,
the ventral or front face of the artefact pointing upwards, and the dorsal or
rear face pointing downwards. This bottom view displays the Z-plane vertically
and the X-plane horizontally, allowing for the simultaneous orthogonal
visualization of both the bilateral left (- X) and right (+ X) sides of the
artefact as well as the bifacial ventral (+ Z) and dorsal (-Z) faces. As the
reference mesh is symmetrical in both bilateral and bifacial aspects, each of
its quadrants are equivalent in volume occupying 1/4 the total volume of the
reference mesh. In bottom view, QI contains the ventral-right quadrant, QII the
dorsal-right quadrant, QIII the dorsal-left quadrant, and QIV the ventral left
quadrant.

To begin, the artefact is bifucated into quadrants by collapsing the artefact
mesh around the X and Z planes in each quadrant. From this perspetive bilateral
or left/right mirror asymmetry can be calculated by isolating opposing volumes
across the plane of reflective symmtery, in bilateral case the Z-plane. To
isolate the artefact's volumetric asymmetry in each quadrant, we determine the
difference between the quadrant's reference volume and the artefact's quadrant
(or segment) volume. For an Endo-Sym analysis, the reference mesh's quarant
volume is subtracted from the artefact's quadrant volume. For an Exo-Sym
analysis, the artefact's quadrant volume is subtracted from the reference mesh's
quadrant volume. In both cases, the remainder is the artefact's volumetric
asymmetry for that quadrant. To calculate bilateral mirror (reflective)
asymmetry across the Z-plane, each face must be analyzed separately, as it
contains both a ventral and a dorsal component. During an analysis of bilateral
mirror symmetry, a comparison of the artefact's  is made In the context of a
volumetric ananlysis refers to the difference between the volume of a
particualar artefact quadrant and the volume of the known symmetrical reference
quadrant in relation to the long axis.

Begining with QI (or the dorsal-right quadrant), the volume of the
known-symmetrical reference mesh is subtracted from the volume of the artefact
model to isolate the artefact's QI or dorsal-right asymmetric volume, also
refered to as the artefact's dorsal-right asymmetry, which is contained in the
variable 'bilateral_dorsal_right_asymmetric_volume'. The same process is
performed in quadrants QII, QIII, and QIV to produce equivalent variables
containing their associated volumetric assymetries.

Next a series of bilateral ventral and dorsal asymmetry coefficients are
calculated based on opposing artefact asymmetries. First the
bilateral_ventral_asymmetry_coefficient is calculated by comparing the
ventral_left_asymmetric_volume to the ventral_right_asymmentric_volume then
dividing the samller volume by the larger. This process is repeated across the
Z-plane to calculate the bilateral_dorsal_asymmetry_coefficient, wherein the
dorsal_left_asymmetric_volume is compared to the
dorsal_right_asymmentric_volume, and the samller volume is divided by the larger
to produce the bilateral_dorsal_asymmetry_coefficient.

An equivalent approach is taken to determine the bifacial left and right
asymmetry coefficients, wherein the dorsal_left_asymmetric_volume is compared to
the dorsal_right_asymmentric_volume, and the samller volume is divided by the
larger to produce the bifacial_left_asymmetry_coefficient. Next the
ventral_left_asymmetric_volume is compared to the
ventral_right_asymmentric_volume, and the samller volume is divided by the
larger to produce the bifacial_right_asymmetry_coefficient.

# Bilateral VENTRAL and DORSAL COEFFICIENTS for COEFFICIENTS APPROACH

Evaluating Handaxe Bilateral/Bifacial Mirror Symmetry Along the Primary Axis in
3D

While any solid bifurcated by a plane along its primary axis can be split into
two shapes (in this case left and right halves or ventral and dorsal faces) with
equivalent volumes, importantly both halves may take an infinite number of
shapes while sharing the primary plane of reflection (the long axis), thus
nullifying the claim that correlated volumes of bifurcated stone tools can be
used as a valid warrant in the analyses of mirror symmetry.

For a particular volume to act as a valid warrant for mirror symmetry, it must
reference a known symmetrical volume. To evaluate bilateral and bifacial mirror
symmetries in reference to the long axis (i.e., the Y-axis), this analyzer
generates a known symmetrical mesh within a shared analysis space oriented to
the same axes as the stone tool. The analysis assumes that a bifacially and
bilaterally symmetrical mesh generated to fit within (the Endo-Sym) or encompass
(the Exo-Sym) the artefact provides the necessary symmetrical reference to
warrant a volume-based analysis of mirror (reflection) symmetry, under the
assumption that within the same analysis space and oriented to the same axes,
any volume occupying the space beyond the surface of the Endo-Sym and beneath
the surface of the Exo-Sym is due to asymmetry inherent in the artefact.

Though the analyses are performed with the artefact in plan view (tip up, butt
down), to better visualize an evaluation of bilateral and bifacial mirror
symmetry, the tip of the artefact is rotated around the X-axis by 270 degrees in
the Z-plane to provide a bottom view of the artefact, with the butt centered,
the ventral or front face of the artefact pointing upwards, and the dorsal or
rear face pointing downwards. This bottom view displays the Z-plane vertically
and the X-plane horizontally, allowing for the simultaneous orthogonal
visualization of both the bilateral left (- X) and right (+ X) sides of the
artefact as well as the bifacial ventral (+ Z) and dorsal (-Z) faces. As the
reference mesh is symmetrical in both bilateral and bifacial aspects, each of
its quadrants are equivalent in volume occupying 1/4 the total volume of the
reference mesh. In bottom view, QI contains the ventral-right quadrant, QII the
dorsal-right quadrant, QIII the dorsal-left quadrant, and QIV the ventral-left
quadrant.

To begin, the artefact is bifurcated into quadrants by collapsing its mesh to
the X and Z planes in each quadrant. From this perspective bilateral or
left/right mirror asymmetry can be calculated by isolating opposing volumes
across the plane of reflective symmetry, in the bilateral case the Z-plane. To
isolate the artefact's volumetric asymmetry in each quadrant, we determine the
difference between the quadrant's reference mesh volume and the artefact's
quadrant (referred to in the code as a segment) volume.

For an Endo-Sym analysis, the reference mesh's quadrant volume is subtracted
from the artefact's quadrant volume. For an Exo-Sym analysis, the artefact's
quadrant volume is subtracted from the reference mesh's quadrant volume. In both
cases, the remainder is the artefact's volumetric asymmetry for that quadrant.

To calculate bilateral mirror (reflective) asymmetry across the Z-plane, each
face must be analyzed separately, as bilateral mirror symmetry contains both a
ventral and a dorsal component. During an Endo-Sym analysis of bilateral mirror
symmetry, a comparison of the artefact's remaining asymmetry volume is compared
to the same volume in the opposing quadrant across the plane of symmetry, in
this case the Z plane. First, to determine the ventral contribution to bilateral
mirror symmetry, the artefact's QI ventral-right volumetric asymmetry is
compared its QIV ventral-left counterpart, and the smaller volume is divided by
the larger volume to produce the artefact's ventral asymmetry coefficient. The
same process is repeated for the dorsal contribution to bilateral mirror
symmetry, wherein the remaining asymmetry volumes of the artefact's dorsal-right
and dorsal-left quadrants (QII and QIII) are compared. The smaller asymmetry
volume is then divided by the larger asymmetry volume to produce the artefact's
dorsal asymmetry coefficient. Both the ventral and dorsal coefficients of
bilateral symmetry are displayed in the Results printed to the Terminal at the
end of the analysis. To achieve the final coefficient of bilateral symmetry, the
ventral and dorsal coefficients are combined into various ratios by dividing the
smaller value by the larger value: the arithmetic mean, the weighted average,
the geometric mean, the harmonic mean, and the root-mean-square. The analyst can
then choose the ratio which best fits the current research question.

Bifacial mirror symmetry is determined by substituting the appropriate remaining
artefact asymmetry volumes to calculate bifacial mirror symmetry across the X
plane. First the asymmetry volumes for QI ventral-right quadrant and QII
dorsal-right are compared. The smaller asymmetry volume is divided by the larger
to determine the right-hand coefficient of bifacial symmetry. The left-hand
coefficient of bifacial symmetry is determined by comparing the artefact's
remaining asymmetry volumes located in the QIII dorsal-left and QIV ventral-left
quadrants. The smaller asymmetry volume is divided by the larger asymmetry
volume to determine the artefact's left-hand coefficient of bifacial symmetry.
Both the left- and right-hand coefficients of bifacial symmetry are displayed in
the Results printed to the Terminal at the end of the analysis. To achieve the
final coefficient of bifacial symmetry, the left- and right-hand coefficients
are combined into the same five ratios offered for bilateral coefficient of
symmetry, with the analyst choosing the ratio appropriate to the research
question.

For Exo-Sym analyses, the equations are modified to achieve an equivalent ratio
between 0 and 1 as the final coefficient of bilateral/bifacial symmetry
described above. Namely, the artefact's asymmetry is now represented by
subtracting the smaller volume of the artefact's segment quadrant from the
larger volume of the Exo-Sym encompassing it. As the Exo-Sym represents the
smallest bilaterally/bifacially symmetrical mesh that can encompass the
artefact, any differences in volume are attributed to asymmetries inherent in
the artefact.

"""

import dataclasses

from . import Scale
from . import ConstructedResult
from . import ArtefactSegments


def calculate_asymmetry_coefficient_of(
    vol_1: float,
    vol_2: float,
) -> float:
    """
    This function determines the value of asymmetry_coefficients by evaluating
    the relationship between two asymmetric_volumes.
    """

    asymmetry_coefficient = min(vol_1, vol_2) / max(vol_1, vol_2)
    return asymmetry_coefficient


@dataclasses.dataclass
class MirrorSymmetry:
    """
    The MirrorSymmetry class uses the current values in asymmetry_coefficient_1
    and asymmetry_coefficient_2 to calculate_bilateral/bifacial mirror_symmetry.
    """

    asymmetry_coefficient_1: float
    asymmetry_coefficient_2: float

    # The following properties define the functions used to calcuate 5 different
    # ratios representing the final bilateral/bifacial_mirror_symmetry
    # coefficients.

    @property
    def arithmetic_mean(self):
        """
        The arithmetic_mean() function returns the arithmetic mean of the two
        asymmetry_coefficients as the final bilateral/bifacial_mirror_symmetry
        coefficient.
        """
        return (self.asymmetry_coefficient_1 + self.asymmetry_coefficient_2) / 2

    @property
    def weighted_average(self):
        """
        The weighted_average() function returns the weighted average of the two
        asymmetry_coefficients as the final bilateral/bifacial_mirror_symmetry
        coefficient.

        This ratio is appropraite if one asymmetry_coefficient or another is
        deemed more significant or reliable for an analysis of symmetry.
        """
        return (
            (self.asymmetry_coefficient_1 * 1) + (self.asymmetry_coefficient_2 * 1)
        ) / (1 + 1)

    @property
    def geometric_mean(self):
        """
        The geometric_mean() function returns the geometric mean of the two
        asymmetry_coefficients.

        This ratio is appropriate if one asymmetry_coefficient varies in scale
        over the other asymmetry_coefficient.
        """
        return (
            abs(self.asymmetry_coefficient_1) * abs(self.asymmetry_coefficient_2)
        ) ** 0.5

    @property
    def harmonic_mean(self):
        """
        The harmonic_mean() function returns the harmonic mean of the two
        asymmetry_coefficients.

        This ratio is sensitive to smaller values and may be useful if smaller
        values of symmetry are important to your analysis.
        """
        return 2 / (
            (1 / self.asymmetry_coefficient_1) + (1 / self.asymmetry_coefficient_2)
        )

    @property
    def rms(self):
        """
        The rms() function returns the root-mean-square of the two
        asymmetry_coefficients.

        This ratio gives more weight to larger values of symmetry if they are
        important for your analysis.
        """
        return (
            ((self.asymmetry_coefficient_1**2) + (self.asymmetry_coefficient_2**2)) / 2
        ) ** 0.5


@dataclasses.dataclass
class ArtefactAnalysis:
    constructed_result: (
        ConstructedResult  # Attribute storing information about constructed result
    )
    artefact_segments: (
        ArtefactSegments  # Attribute storing information about artefact segments
    )

    def __post_init__(self):
        """
        Post-initialization method to calculate additional attributes after the
        object is created.
        """

        # Calculating the volume of the constructed reference mesh and storing
        # it in a variable.
        self.reference_mesh_volume: float = self.constructed_result.mesh.volume()

        # Calculating the original artefact model volume and storing it in a
        # variable.
        self.original_artefact_volume: float = self.artefact_segments.artefact.volume()

        # HEMISPHERIC APPROACH

        # Calculating half of the constructed reference mesh volume and storing
        # it in a variable.
        self.diagnostic_hemisphere_volume: float = self.reference_mesh_volume / 2

        # Calculating the dorsal asymmetric volume and storing it in a variable.
        self.dorsal_asymmetric_volume: float = (
            self.artefact_segments.dorsal.volume() - self.diagnostic_hemisphere_volume
        )

        # Calculating the ventral asymmetric volume and storing it in a
        # variable.
        self.ventral_asymmetric_volume: float = (
            self.artefact_segments.ventral.volume() - self.diagnostic_hemisphere_volume
        )
        # Calculating the left asymmetric volume and storing it in a variable.
        self.left_asymmetric_volume: float = (
            self.artefact_segments.left.volume() - self.diagnostic_hemisphere_volume
        )
        # Calculating the right asymmetric volume and storing it in a variable.
        self.right_asymmetric_volume: float = (
            self.artefact_segments.right.volume() - self.diagnostic_hemisphere_volume
        )
        # Calculating the bilateral asymmetric volume and storing it in a
        # variable.
        self.bilateral_asymmetric_volume: float = (
            self.left_asymmetric_volume - self.right_asymmetric_volume
        )
        # Calculating the bifacial asymmetric volume and storing it in a
        # variable.
        self.bifacial_asymmetric_volume: float = (
            self.ventral_asymmetric_volume - self.dorsal_asymmetric_volume
        )

        # QUADRANT APPROACH

        # Calculating the volume of a single reference mesh quadrant and storing
        # it in a variable.
        self.reference_mesh_quadrant_volume: float = self.reference_mesh_volume / 4

        # Checking if the constructed mesh volume is less (i.e., an Endo_Sym)
        # than the total artefact segments volume.
        if self.reference_mesh_volume < self.artefact_segments.volume:
            # EQUATIONS FOR AN ENDO-SYM ANALYSIS

            # Calculating the omega coefficient of global radial symmetry
            # referencing the Y-axis and storing it in a variable.

            self.radial_symmetry: float = (
                self.reference_mesh_volume / self.artefact_segments.volume
            )

            # Calculating the QII dorsal-right asymmetric volume and storing it
            # in a variable.
            self.dorsal_right_asymmetric_volume: float = (
                self.artefact_segments.dorsal_right.volume()
                - self.reference_mesh_quadrant_volume
            )

            # Calculating the QIII dorsal-left asymmetric volume and storing it
            # in a variable.
            self.dorsal_left_asymmetric_volume: float = (
                self.artefact_segments.dorsal_left.volume()
                - self.reference_mesh_quadrant_volume
            )

            # Calculating the QI ventral-right asymmetric volume and storing it
            # in a variable.
            self.ventral_right_asymmetric_volume: float = (
                self.artefact_segments.ventral_right.volume()
                - self.reference_mesh_quadrant_volume
            )

            # Calculating the QIV ventral-left asymmetric volume and storing it
            # in a variable.
            self.ventral_left_asymmetric_volume: float = (
                self.artefact_segments.ventral_left.volume()
                - self.reference_mesh_quadrant_volume
            )

        else:
            # EQUATIONS FOR AN EXO-SYM ANALYSIS

            # If the constructed mesh volume is greater than or equal to the
            # total artefact segments volume (i.e., an Exo-Sym mesh).
            self.radial_symmetry: float = (
                self.artefact_segments.volume / self.reference_mesh_volume
            )

            # Calculating the QII dorsal-right asymmetric volume and storing it
            # in a variable.
            self.dorsal_right_asymmetric_volume: float = (
                self.reference_mesh_quadrant_volume
                - self.artefact_segments.dorsal_right.volume()
            )
            # Calculating the QIII dorsal-left asymmetric volume and storing it
            # in a variable.
            self.dorsal_left_asymmetric_volume: float = (
                self.reference_mesh_quadrant_volume
                - self.artefact_segments.dorsal_left.volume()
            )
            # Calculating the QI ventral-right asymmetric volume and storing it
            # in a variable.
            self.ventral_right_asymmetric_volume: float = (
                self.reference_mesh_quadrant_volume
                - self.artefact_segments.ventral_right.volume()
            )
            # Calculating the QIV ventral-left asymmetric volume and storing it
            # in a variable.
            self.ventral_left_asymmetric_volume: float = (
                self.reference_mesh_quadrant_volume
                - self.artefact_segments.ventral_left.volume()
            )

        # Calculating the absolute difference between QI ventral-right and QIV
        # ventral-left asymmetric volumes and storing it in a variable.
        self.bilateral_ventral_asymmetric_volume_difference: float = abs(
            self.ventral_right_asymmetric_volume - self.ventral_left_asymmetric_volume
        )
        # Calculating the absolute difference between QII dorsal-right and QIII
        # dorsal-left asymmetric volumes and storing it in a variable.
        self.bilateral_dorsal_asymmetric_volume_difference: float = abs(
            self.dorsal_right_asymmetric_volume - self.dorsal_left_asymmetric_volume
        )
        # Calculating the absolute difference between QIII dorsal-left and QIV
        # ventral-left asymmetric volumes and storing it in a variable.
        self.bifacial_left_asymmetric_volume_difference: float = abs(
            self.dorsal_left_asymmetric_volume - self.ventral_left_asymmetric_volume
        )
        # Calculating the absolute difference between QII dorsal-right and QI
        # ventral-right asymmetric volumes and storing it in a variable.
        self.bifacial_right_asymmetric_volume_difference: float = abs(
            self.dorsal_right_asymmetric_volume - self.ventral_right_asymmetric_volume
        )
        # Calculating the bilateral ventral asymmetry coefficient using a custom
        # function and storing it in a variable.
        self.bilateral_ventral_asymmetry_coefficient: float = (
            calculate_asymmetry_coefficient_of(
                vol_1=self.ventral_left_asymmetric_volume,
                vol_2=self.ventral_right_asymmetric_volume,
            )
        )
        # Calculating the bilateral dorsal asymmetry coefficient using a custom
        # function and storing it in a variable.
        self.bilateral_dorsal_asymmetry_coefficient: float = (
            calculate_asymmetry_coefficient_of(
                vol_1=self.dorsal_left_asymmetric_volume,
                vol_2=self.dorsal_right_asymmetric_volume,
            )
        )
        # Calculating the bifacial left asymmetry coefficient using a custom
        # function and storing it in a variable.
        self.bifacial_left_asymmetry_coefficient: float = (
            calculate_asymmetry_coefficient_of(
                vol_1=self.dorsal_left_asymmetric_volume,
                vol_2=self.ventral_left_asymmetric_volume,
            )
        )
        # Calculating the bifacial right asymmetry coefficient using a custom
        # function and storing it in a variable.
        self.bifacial_right_asymmetry_coefficient: float = (
            calculate_asymmetry_coefficient_of(
                vol_1=self.dorsal_right_asymmetric_volume,
                vol_2=self.ventral_right_asymmetric_volume,
            )
        )
        # Creating a bilateral MirrorSymmetry object with bilateral ventral
        # asymmetry coefficient and bilateral dorsal asymmetry coefficient.
        self.bilateral_mirror_symmetry: MirrorSymmetry = MirrorSymmetry(
            asymmetry_coefficient_1=self.bilateral_ventral_asymmetry_coefficient,
            asymmetry_coefficient_2=self.bilateral_dorsal_asymmetry_coefficient,
        )
        # Creating a bifacial MirrorSymmetry object with bifacial left asymmetry
        # coefficient and bifacial right asymmetry coefficient.
        self.bifacial_mirror_symmetry: MirrorSymmetry = MirrorSymmetry(
            asymmetry_coefficient_1=self.bifacial_left_asymmetry_coefficient,
            asymmetry_coefficient_2=self.bifacial_right_asymmetry_coefficient,
        )
        # VOLUMETRIC REFINEMENT: Calculating the volumetric refinement ratio by
        # dividing artefact volume by surface area and storing it in a variable.
        # self.refinement_volumetric: float = self.artefact_segments.volume / self.artefact_data.oriented_artefact.area() area

        # LINEAR REFINEMENT: Defining variables for minimum and maximum breadth,
        # length, and thickness values.

        self.breadth_min: float  # Minimum breadth value
        self.breadth_max: float  # Maximum breadth value
        self.length_min: float  # Minimum length value
        self.length_max: float  # Maximum length value
        self.thickness_min: float  # Minimum thickness value
        self.thickness_max: float  # Maximum thickness value

        # Extracting minimum and maximum values for breadth, length, and
        # thickness from artefact bounds.
        (
            self.breadth_min,
            self.breadth_max,
            self.length_min,
            self.length_max,
            self.thickness_min,
            self.thickness_max,
        ) = self.artefact_segments.artefact.bounds()

        # Calculating the absolute difference between maximum and minimum
        # breadth values and storing it in a variable.
        breadth: float = abs((self.breadth_max - self.breadth_min))
        # length = abs((length_max - length_min)) Calculating the absolute
        # difference between maximum and minimum thickness values and storing it
        # in a variable.
        thickness: float = abs((self.thickness_max - self.thickness_min))

        # Calculating the linear refinement ratio by dividing thickness by breadth and
        # storing it in a variable.
        self.refinement_linear: float = thickness / breadth

    def to_csv_data_dict(self):
        """
        This method constructs and returns a dictionary where each key-value
        pair represents a specific metric related to an object for exporting or
        analyzing an artefact's geometric, symmetry, and asymmetry properties.
        Each line inside the method's return statement creates a key-value pair
        in the dictionary, where the key is a string describing the data, and
        the value is formatted based on the object's attributes.
        """
        return {
            # This line calculates the reference volume of the mesh and
            # multiplies it by a scaling factor defined by Scale.VOLUME. The
            # result is formatted as a string with four decimal places. This
            # represents the total volume of the reference mesh.
            "Ref_Volume(mm^3)": f"{self.reference_mesh_volume * Scale.VOLUME:0.4f}",
            # This line calculates the volume of a quadrant (1/4) of the
            # reference mesh to a string with four decimal places for more
            # detailed analysis.
            "Ref_Quad_Volume(mm^3)": f"{self.reference_mesh_quadrant_volume:0.4f}",
            # This outputs the omega coefficient of symmetry of the object as a
            # string with four decimal places as a measure of the overall radial
            # symmetry (in reference to the Y-axis) of the artefact.
            "Radial_Symmetry": f"{self.radial_symmetry:0.4f}",
            # This outputs the arithmetic mean of the bilateral mirror symmetry
            # (BLMS) as a string, formatted to four decimal places.
            "Mean_BLMS": f"{self.bilateral_mirror_symmetry.arithmetic_mean:0.4f}",
            # This outputs the coefficient of ventral bilateral asymmetry,
            # converting it into a string with four decimal places. This value
            # represents the asymmetry on the ventral face of the artefact.
            "Ventral_BLMS": f"{self.bilateral_ventral_asymmetry_coefficient:0.4f}",
            # This outputs the coefficient of dorsal bilateral asymmetry,
            # converting it into a string with four decimal places. This value
            # represents the asymmetry on the dorsal face of the artefact.
            "Dorsal_BLMS": f"{self.bilateral_dorsal_asymmetry_coefficient:0.4f}",
            # This outputs the arithmetic mean of the bifacial mirror symmetry
            # (BFMS) as a string, formatted to four decimal places.
            "Mean_BFMS": f"{self.bifacial_mirror_symmetry.arithmetic_mean:0.4f}",
            # This outputs the coefficient of left-hand bifacial asymmetry,
            # converting it into a string with four decimal places. This value
            # represents the asymmetry on the left-hand side of the artefact.
            "Left_BFMS": f"{self.bifacial_left_asymmetry_coefficient:0.4f}",
            # This outputs the coefficient of right-hand bifacial asymmetry,
            # converting it into a string with four decimal places. This value
            # represents the asymmetry on the right-hand side of the artefact.
            "Right_BFMS": f"{self.bifacial_right_asymmetry_coefficient:0.4f}",
            # This line multiplies a refinement ratio by a scaling factor
            # defined by Scale.LENGTH, formatting the result to four decimal
            # places as a string, representing the thinness or 'refinement' of
            # the artefact.
            # "Volumetric_Refinement": f"{self.refinement_volumetric * Scale.VOLUME:0.4f}",
            # This line multiplies a volumetric refinement ratio by a scaling factor
            # defined by Scale.VOLUME, formatting the result to four decimal
            # places as a string, representing the thinness or 'refinement' of
            # the artefact.
            "Linear_Refinement": f"{self.refinement_linear * Scale.LENGTH:0.4f}",
            # This converts the artefact's ventral-right asymmetric volume to a
            # string without specifying a format. It represents the asymmetric
            # volume of QI or the ventral-right quadrant of the artefact.
            "Ventral_Right_Asym_Volume(mm^3)": f"{self.ventral_right_asymmetric_volume:.4f}",
            # This converts the artefact's dorsal-right asymmetric volume to a
            # string without specifying a format. It represents the asymmetric
            # volume of QII or the dorsal-right quadrant of the artefact.
            "Dorsal_Right_Asym_Volume(mm^3)": f"{self.dorsal_right_asymmetric_volume:.4f}",
            # This converts the artefact's dorsal-left asymmetric volume to a
            # string without specifying a format. It represents the asymmetric
            # volume of QIII or the dorsal-left quadrant of the artefact.
            "Dorsal_Left_Asym_Volume(mm^3)": f"{self.dorsal_left_asymmetric_volume:.4f}",
            # This converts the artefact's ventral-left asymmetric volume to a
            # string without specifying a format. It represents the asymmetric
            # volume of QIV or the ventral-left quadrant of the artefact.
            "Ventral_Left_Asym_Volume(mm^3)": f"{self.ventral_left_asymmetric_volume:.4f}",
        }

    def to_volume_data_dict(self):
        return {
            "Artefact_Volume": self.artefact_segments.volume * Scale.VOLUME,
            "Length_Max": self.length_max * Scale.LENGTH,
            "Length_Min": self.length_min * Scale.LENGTH,
            "Thickness_Max": self.thickness_max * Scale.LENGTH,
            "Thickness_Min": self.thickness_min * Scale.LENGTH,
            "Breadth_Max": self.breadth_max * Scale.LENGTH,
            "Breadth_Min": self.breadth_max * Scale.LENGTH,
            "Volume_QI": self.artefact_segments.ventral_right.volume() * Scale.VOLUME,
            "Volume_QII": self.artefact_segments.dorsal_right.volume() * Scale.VOLUME,
            "Volume_QIII": self.artefact_segments.dorsal_left.volume() * Scale.VOLUME,
            "Volume_QIV": self.artefact_segments.ventral_left.volume() * Scale.VOLUME,
        }



from . import OrientationMethod

import vedo
import trimesh
import numpy as np


def identify_tip_vertices(artefact: vedo.Mesh) -> np.ndarray:
    # get the furthest point from the centre of mass

    vertices = artefact.vertices
    centre = get_volumetric_centre(artefact)

    distance_to_centre_of_mass = np.sqrt(np.sum((vertices - centre) ** 2, axis=1))

    tip_index_list = np.argmax(distance_to_centre_of_mass)

    tip_vertices = vertices[tip_index_list]

    return tip_vertices


def get_end_points(artefact: vedo.Mesh) -> tuple[np.ndarray, np.ndarray]:
    vertices = artefact.vertices

    tip_vertices = identify_tip_vertices(artefact=artefact)

    vertices = artefact.vertices

    distance_to_tip = np.sqrt(np.sum((vertices - tip_vertices) ** 2, axis=1))

    butt_index_list = np.argmax(distance_to_tip)

    butt_vertices = vertices[butt_index_list]

    return tip_vertices, butt_vertices


def get_volumetric_centre(artefact: vedo.Mesh):
    points = artefact.vertices
    faces = artefact.cells

    mesh = trimesh.Trimesh(vertices=points, faces=faces)

    return mesh.center_mass


def get_centre_of_point_cloud(artefact: vedo.Mesh):
    return artefact.center_of_mass()


def move_tip_to_origin(artefact: vedo.Mesh) -> vedo.Mesh:
    # get the furthest point from the centre of mass

    temp = artefact.clone()

    artefacts = []
    artefacts.append(artefact.clone().c("red").alpha(0.5))

    vertices = artefact.vertices
    centre = get_volumetric_centre(artefact)

    distance_to_centre_of_mass = np.sqrt(np.sum((vertices - centre) ** 2, axis=1))

    tip_index_list = np.argmax(distance_to_centre_of_mass)

    tip_vertices = vertices[tip_index_list]

    temp.align_with_landmarks(
        [tip_vertices, centre],
        [[0, 0, 0], [0, -distance_to_centre_of_mass.max(), 0]],
        rigid=True,
    )

    artefacts.append(temp.clone().c("green").alpha(0.5))
    if False:
        plt = vedo.Plotter()
        plt.show(artefacts, axes=True)

    artefact = temp.clone()

    return artefact


def orient_tip_to_centre_of_mass(artefact: vedo.Mesh) -> vedo.Mesh:
    return_artefact = artefact.clone()

    tip_mean, butt_mean = get_end_points(artefact=return_artefact)

    origin = np.asarray((0, 0, 0))

    com = get_volumetric_centre(return_artefact)

    r = np.sqrt(((tip_mean - com) ** 2).sum())

    new_com = [0, -r, 0]
    return_artefact.align_with_landmarks([tip_mean, com], [origin, new_com])

    if False:
        xt, yt, zt = tip_mean
        # xb, yb, zb = com

        p0_y = yt + 50
        p1_y = butt_mean[1] - 50
        axis_line = vedo.Line(p0=(0, p0_y, 0), p1=(0, p1_y, 0))
        plt = vedo.Plotter()
        plt.show(return_artefact.c("blue"), artefact.alpha(0.3), axis_line, axes=True)

    return return_artefact.clone()


def orient_tip_to_centre_of_hollow_mesh(artefact: vedo.Mesh) -> vedo.Mesh:
    return_artefact = artefact.clone()

    tip_mean, butt_mean = get_end_points(artefact=return_artefact)

    origin = np.asarray((0, 0, 0))

    com = get_centre_of_point_cloud(return_artefact)

    r = np.sqrt(((tip_mean - com) ** 2).sum())

    new_com = [0, -r, 0]
    return_artefact.align_with_landmarks([tip_mean, com], [origin, new_com])

    if False:
        xt, yt, zt = tip_mean
        # xb, yb, zb = com

        p0_y = yt + 50
        p1_y = butt_mean[1] - 50
        axis_line = vedo.Line(p0=(0, p0_y, 0), p1=(0, p1_y, 0))
        plt = vedo.Plotter()
        plt.show(return_artefact.c("blue"), artefact.alpha(0.3), axis_line, axes=True)

    return return_artefact.clone()


def orient_tip_to_butt(artefact: vedo.Mesh) -> vedo.Mesh:
    return_artefact = artefact.clone()

    tip_mean, butt_mean = get_end_points(artefact=return_artefact)

    origin = np.asarray((0, 0, 0))

    r = np.sqrt(((tip_mean - butt_mean) ** 2).sum())

    new_butt = [0, -r, 0]
    return_artefact.align_with_landmarks([tip_mean, butt_mean], [origin, new_butt])

    return return_artefact.clone()


def orient_to_fitline(artefact: vedo.Mesh) -> vedo.Mesh:
    # start with tip at origin for correct orientation
    return_artefact = move_tip_to_origin(artefact=artefact).clone()

    orienting_artefact = return_artefact.clone().decimate(fraction=0.1, method="quad")

    vertices = orienting_artefact.vertices
    # TODO: handle crash with non decimated artefact this uses lots of memory
    # and takes a long time if the following line causes a crash try decimating
    # the mesh before running the analysis
    fitline = vedo.fit_line(vertices)

    points = return_artefact.intersect_with_line(fitline)
    a = points[points[:, 1].argsort()][::-1]

    com = return_artefact.center_of_mass()

    tip_intersect = a[0]
    butt_intersect = a[-1]

    rt = np.sqrt(((tip_intersect - com) ** 2).sum())
    rb = np.sqrt(((butt_intersect - com) ** 2).sum())

    if rb > rt:
        tip_intersect, butt_intersect = butt_intersect, tip_intersect

    r = np.sqrt(((tip_intersect - butt_intersect) ** 2).sum())

    new_butt = [0, -r, 0]
    return_artefact.align_with_landmarks(
        [tip_intersect, butt_intersect], [(0, 0, 0), new_butt]
    )

    if False:
        plt = vedo.Plotter()
        plt.show(return_artefact.c("blue"), artefact, axes=True)

    return return_artefact.clone()


def align_thickness_with_x_axis(artefact: vedo.Mesh) -> vedo.Mesh:
    xmin, xmax, ymin, ymax, zmin, zmax = artefact.bounds()

    artefacts = []
    artefacts.append(artefact.clone().c("red"))

    vertices = artefact.vertices
    distance_to_y_axis = (vertices[:, 0] ** 2 + vertices[:, 2] ** 2) ** 0.5
    opposites = vertices[:, 2]
    adjacents = vertices[:, 0]
    angles_from_x = np.arctan(opposites / adjacents)

    furthest_point_index = np.argwhere(distance_to_y_axis == max(distance_to_y_axis))

    this_angle = angles_from_x[furthest_point_index.max()]

    artefact.rotate_y(this_angle, rad=True)
    artefacts.append(artefact.clone().c("yellow"))

    vertices = artefact.vertices

    max_x_value = vertices[:, 0].max()
    min_x_value = vertices[:, 0].min()

    max_x_list = vertices[vertices[:, 0] == max_x_value]
    min_x_list = vertices[vertices[:, 0] == min_x_value]

    r = 0
    x_min = None
    x_max = None

    for min_x in min_x_list:
        for max_x in max_x_list:
            this_r = abs(max_x[0] - min_x[0])
            if this_r > r:
                x_min = min_x
                x_max = max_x

    adjacent = x_max[0] - x_min[0]
    opposite = x_max[2] - x_min[2]

    widest_points_average_y_value = (x_max[1] + x_min[1]) / 2
    theta = np.arctan(opposite / adjacent)

    artefact.rotate_y(theta, rad=True)

    # Setting this to true will orient artefact using widest part as closest to
    # butt
    if False:
        artefact_length = abs(ymax - ymin)
        artefact_half_way = -artefact_length / 2

        if artefact_half_way <= widest_points_average_y_value:
            artefact.rotate_z(np.pi, rad=True)

    artefacts.append(artefact.clone().c("green"))

    return artefact.clone()


def orient_artefact(
    artefact: vedo.Mesh,
    artefact_name: str,
    orientation_method: OrientationMethod,
) -> vedo.Mesh:
    artefacts = []

    # artefacts.append(artefact.clone().c("red"))

    if orientation_method is OrientationMethod.ORIENT_TIP_TO_BUTT:
        artefact = orient_tip_to_butt(artefact=artefact)
        artefacts.append(artefact.clone().c("yellow"))

    elif orientation_method is OrientationMethod.ORIENT_TIP_TO_COM:
        artefact = orient_tip_to_centre_of_mass(artefact=artefact)
        artefacts.append(artefact.clone().c("yellow"))

    elif orientation_method is OrientationMethod.ORIENT_TIP_TO_COHM:
        artefact = orient_tip_to_centre_of_hollow_mesh(artefact=artefact)
        artefacts.append(artefact.clone().c("yellow"))

    elif orientation_method is OrientationMethod.ORIENT_TO_FITLINE:
        artefact = orient_to_fitline(artefact=artefact)
        artefacts.append(artefact.clone().c("yellow"))

    artefact = align_thickness_with_x_axis(artefact=artefact)
    artefacts.append(artefact.clone().c("blue"))

    centre = get_volumetric_centre(artefact)

    artefact = artefact.shift(dy=-centre[1]).clone()

    artefacts.append(artefact.clone().c("green"))

    return artefact.clone()



import enum
import vedo
import numpy as np


class SegmentationMethod(enum.Enum):
    COLLAPSE_TO_PLANE = 1  # Creates a flange artefact on quadrants
    COLLAPSE_TO_POINT = 2  # Creates a set of triangular artefacts on plane faces


class ArtefactSegments:
    def segment_mesh_dorsal(
        self,
        mesh: vedo.Mesh,
        segmentation_method: SegmentationMethod,
    ) -> vedo.Mesh:
        points = np.asarray(mesh.vertices)
        new_points = points.copy()
        match segmentation_method:
            case SegmentationMethod.COLLAPSE_TO_PLANE:
                np.minimum(new_points[:, 2], 0, out=new_points[:, 2])
            case SegmentationMethod.COLLAPSE_TO_POINT:
                new_points[new_points[:, 2] > 0] = 0
        new_mesh = vedo.Mesh((new_points, mesh.cells))
        return new_mesh

    def segment_mesh_ventral(
        self,
        mesh: vedo.Mesh,
        segmentation_method: SegmentationMethod,
    ) -> vedo.Mesh:
        points = np.asarray(mesh.vertices)
        new_points = points.copy()
        match segmentation_method:
            case SegmentationMethod.COLLAPSE_TO_PLANE:
                np.maximum(new_points[:, 2], 0, out=new_points[:, 2])
            case SegmentationMethod.COLLAPSE_TO_POINT:
                new_points[new_points[:, 2] < 0] = 0
        new_mesh = vedo.Mesh((new_points, mesh.cells))
        return new_mesh

    def segment_mesh_left(
        self,
        mesh: vedo.Mesh,
        segmentation_method: SegmentationMethod,
    ) -> vedo.Mesh:
        points = np.asarray(mesh.vertices)
        new_points = points.copy()
        match segmentation_method:
            case SegmentationMethod.COLLAPSE_TO_PLANE:
                np.minimum(new_points[:, 0], 0, out=new_points[:, 0])
            case SegmentationMethod.COLLAPSE_TO_POINT:
                new_points[new_points[:, 0] > 0] = 0
        new_mesh = vedo.Mesh((new_points, mesh.cells))
        return new_mesh

    def segment_mesh_right(
        self,
        mesh: vedo.Mesh,
        segmentation_method: SegmentationMethod,
    ) -> vedo.Mesh:
        points = np.asarray(mesh.vertices)
        new_points = points.copy()
        match segmentation_method:
            case SegmentationMethod.COLLAPSE_TO_PLANE:
                np.maximum(new_points[:, 0], 0, out=new_points[:, 0])
            case SegmentationMethod.COLLAPSE_TO_POINT:
                new_points[new_points[:, 0] < 0] = 0
        new_mesh = vedo.Mesh((new_points, mesh.cells))
        return new_mesh

    def __init__(self, artefact: vedo.Mesh):
        # SEGMENTATION_METHOD = SegmentationMethod.COLLAPSE_TO_POINT
        SEGMENTATION_METHOD = SegmentationMethod.COLLAPSE_TO_PLANE

        self.artefact = artefact

        self.volume = artefact.volume()
        self.points = np.asarray(artefact.vertices)
        self.faces = artefact.cells

        self.dorsal = self.segment_mesh_dorsal(
            mesh=artefact,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.ventral = self.segment_mesh_ventral(
            mesh=artefact,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.left = self.segment_mesh_left(
            mesh=artefact,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.right = self.segment_mesh_right(
            mesh=artefact,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.dorsal_right = self.segment_mesh_right(
            mesh=self.dorsal,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.dorsal_left = self.segment_mesh_left(
            mesh=self.dorsal,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.ventral_right = self.segment_mesh_right(
            mesh=self.ventral,
            segmentation_method=SEGMENTATION_METHOD,
        )

        self.ventral_left = self.segment_mesh_left(
            mesh=self.ventral,
            segmentation_method=SEGMENTATION_METHOD,
        )



import dataclasses

from . import QuadrantAttributes
from . import Quadrant


@dataclasses.dataclass
class MeasurementAttributes:
    full_angle: float
    measurement_angle: float
    quadrant_angle: float
    quadrant_data: QuadrantAttributes


class MeasurementData:
    def __init__(
        self,
        quadrant: Quadrant,
        segment_count: int,
    ):
        self.quadrant = quadrant
        self.quadrant_attributes = quadrant.value

        self.segment_count = segment_count
        self.segment_angle = 90 / (segment_count + 1)
        self.angles = [0.0]
        self.angles += list(
            [angle * self.segment_angle for angle in range(1, self.segment_count + 1)]
        )
        self.angles.append(90)
        self.angle_iter = iter(self.angles)

    def __iter__(self):
        self.angle_iter = iter(self.angles)
        return self

    def __next__(self) -> MeasurementAttributes:
        angle = next(self.angle_iter)

        match angle:
            case 0:
                full_angle = self.quadrant_attributes.start_angle
                # measurement angles 0 and 360 should yield the same result,
                # so we only need to measure once.
                if full_angle == 360:
                    measurement_angle = 0
                else:
                    measurement_angle = full_angle

                return_val = MeasurementAttributes(
                    full_angle=full_angle,
                    measurement_angle=measurement_angle,
                    quadrant_angle=angle,
                    quadrant_data=self.quadrant_attributes,
                )
            case 90:
                full_angle = self.quadrant_attributes.end_angle
                measurement_angle = full_angle
                return_val = MeasurementAttributes(
                    full_angle=full_angle,
                    measurement_angle=measurement_angle,
                    quadrant_angle=90,
                    quadrant_data=self.quadrant_attributes,
                )

            case _:
                full_angle = (
                    self.quadrant_attributes.start_angle
                    + angle * self.quadrant_attributes.angular_direction.value
                )
                measurement_angle = full_angle

                return_val = MeasurementAttributes(
                    full_angle=full_angle,
                    measurement_angle=measurement_angle,
                    quadrant_angle=angle,
                    quadrant_data=self.quadrant_attributes,
                )
        return return_val



from . import OrientationMethod
from . import SliceData
from . import SampleType

import numpy as np
import dataclasses


@dataclasses.dataclass
class ArtefactInfo:
    end_points: tuple[np.ndarray, np.ndarray]

    name: str | None = None
    original_slice_data: list[SliceData] = dataclasses.field(default_factory=list)
    orientation_method: OrientationMethod | None = None

    samples: np.ndarray = dataclasses.field(
        default_factory=lambda: np.empty(shape=0, dtype=SampleType).view(np.recarray)
    )



"""Code to create meshes from points"""

import numpy as np
from tqdm import tqdm
import vedo

from . import AngularDirection
from . import Quadrant
from . import ConstructedResult
from . import ConstructionData
from ._sample import PointXyzType


def get_point_index(
    points: np.ndarray,
    point: np.ndarray,
) -> int:
    """Finds matching index for entries in points which match the valus in point"""
    match = (points == point).all(axis=1)
    return np.where(match)[0]


def make_segment_faces_np(
    points: np.ndarray,
    slice1: np.ndarray,
    slice2: np.ndarray,
) -> np.ndarray:
    """Builds the faces between the two slices to create a complete ring"""

    faces = []

    quadrant_angles = np.unique(slice1.angle.quadrant_angle)
    quadrant_angles.sort()

    for quadrant_attributes in Quadrant:
        angular_direction = quadrant_attributes.value.angular_direction
        direction = angular_direction == AngularDirection.clockwise
        start_angle = quadrant_attributes.value.start_angle

        quadrant_angle = quadrant_angles[0]

        slice1_point = slice1[
            (slice1.angle.start_angle == start_angle)
            & (slice1.angle.clockwise == direction)
            & (slice1.angle.quadrant_angle == quadrant_angle)
        ]

        actual_slice1_point = slice1_point.point.view(PointXyzType)

        slice1_previous_point_index = get_point_index(
            points=points,
            point=actual_slice1_point,
        )

        slice2_point = slice2[
            (slice2.angle.start_angle == start_angle)
            & (slice2.angle.clockwise == direction)
            & (slice2.angle.quadrant_angle == quadrant_angle)
        ]

        actual_slice2_point = slice2_point.point.view(PointXyzType)

        slice2_previous_point_index = get_point_index(
            points=points,
            point=actual_slice2_point,
        )

        for quadrant_angle in quadrant_angles[1:]:
            slice1_point = slice1[
                (slice1.angle.start_angle == start_angle)
                & (slice1.angle.clockwise == direction)
                & (slice1.angle.quadrant_angle == quadrant_angle)
            ]

            actual_slice1_point = slice1_point.point.view(PointXyzType)

            slice1_point_index = get_point_index(
                points=points,
                point=actual_slice1_point,
            )

            slice2_point = slice2[
                (slice2.angle.start_angle == start_angle)
                & (slice2.angle.clockwise == direction)
                & (slice2.angle.quadrant_angle == quadrant_angle)
            ]

            actual_slice2_point = slice2_point.point.view(PointXyzType)

            slice2_point_index = get_point_index(
                points=points,
                point=actual_slice2_point,
            )

            face1 = np.concatenate(
                (
                    slice1_previous_point_index,
                    slice1_point_index,
                    slice2_previous_point_index,
                )
            )
            faces.append(face1)

            face2 = np.concatenate(
                (
                    slice2_previous_point_index,
                    slice1_point_index,
                    slice2_point_index,
                )
            )
            faces.append(face2)

            slice1_previous_point_index = slice1_point_index
            slice2_previous_point_index = slice2_point_index

    return faces


def make_end_faces_np(
    points: np.ndarray,
    end_point: np.ndarray,
    artefact_slice: np.ndarray,
) -> np.ndarray:
    """Builds faces from the last slice to the end point"""

    faces = []

    end_index = get_point_index(points=points, point=end_point)

    quadrant_angles = np.unique(artefact_slice.angle.quadrant_angle)
    quadrant_angles.sort()

    for quadrant_attributes in Quadrant:
        angular_direction = quadrant_attributes.value.angular_direction
        direction = angular_direction == AngularDirection.clockwise
        start_angle = quadrant_attributes.value.start_angle

        quadrant_angle = quadrant_angles[0]

        point = artefact_slice[
            (artefact_slice.angle.start_angle == start_angle)
            & (artefact_slice.angle.clockwise == direction)
            & (artefact_slice.angle.quadrant_angle == quadrant_angle)
        ]

        actual_point = point.point.view(PointXyzType)

        previous_point_index = get_point_index(
            points=points,
            point=actual_point,
        )

        for quadrant_angle in quadrant_angles[1:]:
            point = artefact_slice[
                (artefact_slice.angle.start_angle == start_angle)
                & (artefact_slice.angle.clockwise == direction)
                & (artefact_slice.angle.quadrant_angle == quadrant_angle)
            ]

            actual_point = point.point.view(PointXyzType)

            point_index = get_point_index(
                points=points,
                point=actual_point,
            )
            face = np.concatenate((end_index, point_index, previous_point_index))
            faces.append(face)
            previous_point_index = point_index

    return faces


def create_mesh_from_point_data(
    construction_data: ConstructionData,
) -> ConstructedResult:
    """Creates a mesh from the point data provided"""

    analysis_method = construction_data.analysis_method
    sample_data = construction_data.sample_data

    tip_mean, butt_mean = construction_data.end_points

    sample_points = sample_data.point.view(PointXyzType)

    sample_lengths = np.unique(sample_data.point.length)

    sample_lengths[::-1].sort()

    # add a point for the tip to points
    sample_points = np.concatenate((sample_points, np.asarray((tip_mean,))))

    # add a point for the butt to points
    sample_points = np.concatenate((sample_points, np.asarray((butt_mean,))))

    # there are duplicated points at the edges of the quadrants
    sample_points = np.unique(sample_points, axis=0)  # remove duplicate points

    # make the faces from the tip to the first slice
    sample_artefact_slice = sample_data[sample_data.point.length == sample_lengths[0]]

    sample_tip_faces = make_end_faces_np(
        points=sample_points,
        end_point=tip_mean,
        artefact_slice=sample_artefact_slice,
    )

    sample_faces = sample_tip_faces

    # make faces for all the slices
    previous_sample_slice = sample_data[sample_data.point.length == sample_lengths[0]]

    desc = f"Synthesising:{analysis_method.text}"
    for i in tqdm(range(1, len(sample_lengths)), desc=desc):
        sample_artefact_slice = (
            sample_points[sample_points[:, 1] == sample_lengths[i]],
        )
        sample_artefact_slice = sample_data[
            sample_data.point.length == sample_lengths[i]
        ]

        # make faces between two slices

        sample_segment_faces = make_segment_faces_np(
            points=sample_points,
            slice1=previous_sample_slice,
            slice2=sample_artefact_slice,
        )

        previous_sample_slice = sample_artefact_slice

        sample_faces = np.concatenate((sample_faces, sample_segment_faces))

    # make faces from last slice to butt

    sample_artefact_slice = sample_data[sample_data.point.length == sample_lengths[-1]]

    sample_butt_faces = make_end_faces_np(
        points=sample_points,
        end_point=butt_mean,
        artefact_slice=sample_artefact_slice,
    )

    sample_faces = np.concatenate((sample_faces, sample_butt_faces))

    sample_faces = np.unique(sample_faces, axis=0)  # remove duplicate faces

    sample_points_vedo = vedo.Points(sample_points)
    sample_mesh = vedo.Mesh([sample_points, sample_faces]).compute_normals()

    if False:
        # The following can be used to create a mesh when pasted into a python
        # script and added to a numpy ndarray print(points)
        for point in sample_points_vedo.vertics:
            print(f"[{point[0]}, {point[1]}, {point[2]}],")
        # print(sample_faces)
        for face in sample_faces:
            print(f"[{face[0]}, {face[1]}, {face[2]}],")

    if False:
        artefact_axes = vedo.Axes(sample_mesh)

        vedo.Plotter().show(
            artefact_axes,
            sample_points_vedo,
            sample_mesh.alpha(0.5),
        )

    sample_constructed_result = ConstructedResult(
        points=sample_points_vedo.clean(),
        mesh=sample_mesh.clean(),
    )

    return sample_constructed_result



"""Contains command line invocation implementation for lithikos"""

import argparse
import pathlib

import vedo

from . import AnalysisMethod
from . import OrientationMethod
from . import SampleParameters
from . import process


ARTEFACT_DATA_FILE_NAME = "artefact_data.csv"

REFERENCE_DATA_FILE_NAME = "reference_data.csv"


def main():
    """Contains command line invocation implementation for lithikos"""

    p = pathlib.Path(".")

    output_root = pathlib.Path("output")
    quadrant_subdir = pathlib.Path("quadrants")

    analysis_method_choices = [method.text for method in AnalysisMethod]
    orientation_method_choices = [method.text for method in OrientationMethod]

    parser = argparse.ArgumentParser(
        prog="lithikos",
        description="""Orients 3D models of stone artefacts, isolates quadrants
        about the long axis of the tool, measures artefact surface area and
        artefact/quadrant volumes and dimensions, samples artefact model geometry,
        synthesises radially symmetrical reference meshes, and delivers oriented
        artefact/quadrant models, symmetrical reference models, and artefact_data.csv
        and reference_data.csv files""",
        epilog="Thank you for using lithikos",
    )

    parser.add_argument(
        "filenames", help="One or more files containing mesh objects in Wavefront (obj) format",
        nargs=argparse.ONE_OR_MORE,
    )

    parser.add_argument(
        "--orientation-method",
        choices=orientation_method_choices,
        nargs=argparse.ONE_OR_MORE,
        default=(OrientationMethod.ORIENT_TIP_TO_COM.text,),
    )

    parser.add_argument(
        "--analysis-method",
        choices=analysis_method_choices,
        nargs=argparse.ONE_OR_MORE,
        default=(AnalysisMethod.ENDO_SYM.text,),
    )

    parser.add_argument(
        "--slice-count",
        type=int,
        default=100,
    )

    parser.add_argument(
        "--point-count",
        type=int,
        default=36,
    )

    parser.add_argument("--tip-spacing", type=float, default=0.001)

    parser.add_argument("--butt-spacing", type=float, default=0.001)

    args = parser.parse_args()

    sample_parameters = SampleParameters(
        slice_count=args.slice_count,
        point_count=args.point_count,
        tip_spacing=args.tip_spacing,
        butt_spacing=args.butt_spacing,
    )

    orientation_methods = [
        OrientationMethod.from_text(method) for method in args.orientation_method
    ]
    analysis_methods = [
        AnalysisMethod.from_text(method) for method in args.analysis_method
    ]

    sample_parameters_list = [sample_parameters]

    for filename in args.filenames:
        artefact_path = pathlib.Path(filename)

        artefact_name = artefact_path.stem

        parent_dir = artefact_path.parent.name

        output_dir = output_root / parent_dir

        output_dir.mkdir(parents=True, exist_ok=True)

        original_artefact = vedo.Mesh(artefact_path.as_posix()).clean()

        artefact_data_file = p / pathlib.Path(ARTEFACT_DATA_FILE_NAME)
        reference_data_file = p / pathlib.Path(REFERENCE_DATA_FILE_NAME)

        process.process(
            original_artefact=original_artefact,
            artefact_name=artefact_name,
            output_dir=output_dir,
            orientation_methods=orientation_methods,
            analysis_methods=analysis_methods,
            sample_parameters_list=sample_parameters_list,
            artefact_data_file=artefact_data_file,
            reference_data_file=reference_data_file,
            quadrant_subdir=quadrant_subdir,
        )


if __name__ == "__main__":
    main()



"""
Functions to analyse data based on the analysis method and produce a set of
symmetrised samples
"""

import copy

from tqdm import tqdm

import numpy as np

from . import ConstructionData
from . import AnalysisMethod
from . import ArtefactInfo


def analyse_artefact_samples(
    artefact_info: ArtefactInfo,
    analysis_method: AnalysisMethod,
) -> ConstructionData:
    """
    Works down the artefact samples, slice by slice and produces a set of
    symmetrised points based on the analysis method
    """

    construction_data = ConstructionData(
        analysis_method=analysis_method,
        end_points=artefact_info.end_points,
    )

    samples = artefact_info.samples

    long_samples: np.recarray = copy.deepcopy(samples).view(np.recarray)

    quadrant_angles = np.unique(long_samples.angle.quadrant_angle)
    quadrant_angles.sort()

    sample_lengths = np.unique(samples.point.length)
    sample_lengths[::-1].sort()  # sort descending
    slice_count = sample_lengths.size

    desc = f"Symmetrising Radial Slices:{analysis_method.text}"
    for i in tqdm(range(slice_count), desc=desc):
        sample_length = sample_lengths[i]
        length_match = long_samples.point.length == sample_length
        for angle in quadrant_angles:
            angle_match = long_samples.angle.quadrant_angle == angle

            points = long_samples.point[angle_match & length_match]
            angles = long_samples.angle[angle_match & length_match]

            # point_magnitudes = np.sqrt(points.breadth**2 + points.thickness**2)
            point_magnitudes = np.hypot(points.breadth, points.thickness)

            match analysis_method:
                case AnalysisMethod.ENDO_SYM:
                    magnitude_to_use = point_magnitudes.min()
                case AnalysisMethod.MEAN_SYM:
                    magnitude_to_use = point_magnitudes.mean()
                case AnalysisMethod.EXO_SYM:
                    magnitude_to_use = point_magnitudes.max()

            new_breadth = magnitude_to_use * np.cos(angle * np.pi / 180)
            new_thickness = magnitude_to_use * np.sin(angle * np.pi / 180)

            long_samples.point.breadth[angle_match & length_match] = (
                new_breadth * angles.breadth_sign
            )
            long_samples.point.thickness[angle_match & length_match] = (
                new_thickness * angles.thickness_sign
            )

    construction_data.sample_data = long_samples
    return construction_data



"""Sample points from a 3d scanned mesh organised in quadrants"""

from tqdm import tqdm

import numpy as np
from numpy.lib.recfunctions import stack_arrays

import vedo

from ._sample import SampleType

from . import AngularDirection
from . import ArtefactInfo
from . import MeasurementData
from .measurement_data import MeasurementAttributes
from . import Quadrant
from . import SliceData
from . import SampleParameters

from . import logger


def get_artefact_y_intersect(
    artefact: vedo.Mesh,
) -> tuple[np.ndarray]:
    """Get the points where the y axis intersects the artefact"""
    (
        _,  # breadth_min,
        _,  # breadth_max,
        length_min,
        length_max,
        _,  # thickness_min,
        _,  # thickness_max,
    ) = artefact.bounds()

    artefact_length = abs(length_max - length_min)

    extention = artefact_length * 0.05

    p0 = (0, length_max + extention, 0)
    p1 = (0, length_min - extention, 0)
    axis_line = vedo.Line(p0=p0, p1=p1)

    if False:
        plt = vedo.Plotter()
        plt.show(artefact.clone().alpha(0.3), axis_line, axes=True)

    points = artefact.intersect_with_line(axis_line)
    points_sorted = points[points[:, 1].argsort()][::-1]

    # TODO: choose which method to use if multiple points are identified at the
    # y intersects

    # this will use external points
    tip_intersect = points_sorted[0]
    butt_intersect = points_sorted[-1]

    # this will use innermost points tip_intersect =
    # points_sorted[points_sorted[:,1] > 0][-1] butt_intersect =
    # points_sorted[points_sorted[:,1] < 0][0]

    return tip_intersect, butt_intersect


def get_intersection_at_angle(
    artefact: vedo.Mesh,
    angle: float,
    magnitude: float,
    origin: tuple[int],
) -> np.ndarray:
    """
    Get the points at the intersection of a line perpendicular to the y axis at
    the angle an origin given
    """

    offset_breadth, offset_length, offset_thickness = origin

    calculated_thickness = np.sin(angle * np.pi / 180) * magnitude
    calculated_breadth = np.cos(angle * np.pi / 180) * magnitude
    p0 = (offset_breadth, offset_length, offset_thickness)
    p1 = (
        offset_breadth + calculated_breadth,
        offset_length,
        offset_thickness + calculated_thickness,
    )
    line = vedo.Line(p0=p0, p1=p1)

    if False:
        plt = vedo.Plotter()
        plt.show(artefact.c("green").alpha(0.1), line.c("red"))

    points = artefact.intersect_with_line(line)

    # TODO add check for multiple or no points being available We only want one
    # point returned from the intersection
    if len(points) == 1:
        point_to_return = points[-1]
    elif len(points) > 1:
        logger.warning(
            "intersect with line at angle: %f origin: %s found %i points",
            angle,
            origin,
            len(points),
        )
        for point in points:
            logger.warning("%s", point)

        if True:  # Use the first point encountered
            point_to_return = points[0]
            logger.warning("using first point as point %s", point_to_return)
        else:  # Use the mean of all points
            point_to_return = np.asanyarray(points).mean(axis=0)
            logger.warning("using mean as point %s", point_to_return)

    else:
        logger.warning(
            "intersect with line at angle: %f origin: %s found %i points",
            angle,
            origin,
            len(points),
        )
        point_to_return = origin
        logger.warning("using slice origin as point %s", point_to_return)

        if False:
            (
                breadth_min,
                breadth_max,
                length_min,
                length_max,
                thickness_min,
                thickness_max,
            ) = artefact.bounds()

            breadth = abs(breadth_max - breadth_min)
            # length = abs(length_max - length_min)
            thickness = abs(thickness_max - thickness_min)

            slice_plane = vedo.Plane(
                pos=origin,
                normal=(0, 1, 0),
                s=(breadth, thickness),
                c="yellow",
                alpha=0.5,
            )
            axis_line = vedo.Line(p0=(0, -200, 0), p1=(0, 200, 0))
            plt = vedo.Plotter()
            plt.show(
                artefact, slice_plane, line.c("red"), axis_line.c("green"), axes=True
            )

    return line, point_to_return


def get_slice_points(
    artefact: vedo.Mesh,
    slice_location: float,
    point_count: int,
    normalise_axis: bool = False,  # see comment below regarding this method
) -> tuple[SliceData, np.ndarray]:
    """Get a list of points for all the angles of a slice"""

    slice_data = SliceData()

    slice_samples = np.empty(0, dtype=SampleType).view(np.recarray)

    (
        breadth_min,
        breadth_max,
        _,  # length_min,
        _,  # length_max,
        thickness_min,
        thickness_max,
    ) = artefact.bounds()

    breadth_limits = abs(
        breadth_max - breadth_min
    )  # make the line length much longer than it needs to be

    slice_origin = (0, slice_location, 0)

    # TODO: consider whether this makes sense This method shifts the centre of
    # each slice before processing Note: this will lose a certain amount of
    # information regarding bi-facial and bi-lateral symmetry
    if normalise_axis:
        slice_data.this_slice = artefact.clone().slice(
            origin=slice_origin, normal=(0, 1, 0)
        )
        slice_data.origin = slice_data.this_slice.center_of_mass()
    else:
        slice_data.this_slice = None
        slice_data.origin = slice_origin

    points_per_quadrant = point_count // 4
    segment_count = points_per_quadrant - 1

    plot_points = []

    entry_list: list[MeasurementAttributes] = []

    for q in Quadrant:
        for entry in MeasurementData(quadrant=q, segment_count=segment_count):
            entry_list.append(entry)

    measurement_angles: list[float] = sorted({m.measurement_angle for m in entry_list})

    for measurement_angle in measurement_angles:
        line, point = get_intersection_at_angle(
            artefact=artefact,
            angle=measurement_angle,
            magnitude=breadth_limits,
            origin=slice_data.origin,
        )
        slice_data.lines.append(line)

        # list of points to log position
        plot_point = vedo.Point(point, r=4)
        plot_points.append(plot_point)

        entry: MeasurementAttributes
        for entry in [
            entry
            for entry in entry_list
            if entry.measurement_angle == measurement_angle
        ]:
            quadrant_data = entry.quadrant_data

            # pylint: disable=too-many-function-args
            this_sample = np.void(0, dtype=SampleType).view(np.recarray)
            this_sample.angle.full_angle = entry.full_angle
            this_sample.angle.quadrant_angle = entry.quadrant_angle
            this_sample.angle.start_angle = quadrant_data.start_angle
            this_sample.angle.clockwise = (
                quadrant_data.angular_direction == AngularDirection.clockwise
            )
            this_sample.angle.breadth_sign = quadrant_data.breadth_sign.value
            this_sample.angle.thickness_sign = quadrant_data.thickness_sign.value

            this_sample.point.breadth = point[0]
            this_sample.point.length = slice_location
            this_sample.point.thickness = point[2]

            slice_samples = stack_arrays(
                (slice_samples, (this_sample,)), asrecarray=True, usemask=False
            )

            # The following code displays the artefact and with a single radial
            # line for each point sample around the current slice. The next
            # plote below displays all points per slice.

        if False:
            artefact_thickness = abs(thickness_min - thickness_max)
            plt = vedo.Plotter()
            plt.camera.SetPosition((0, 0, artefact_thickness))
            plt.camera.SetFocalPoint((0, 0, 0))
            plt.show(artefact.alpha(0.3), line, plot_point, axes=True)

    # The following code displays the Radial Points plot showing the artefact
    # with the point_count located at first slice below the tip. Closing this
    # initial plot opens the same view for each of the slices defined in the
    # slice_data variable.

    if False:  # show radial points per slice
        title = "Radial Points Measured per Slice"
        artefact_thickness = abs(thickness_min - thickness_max)
        plt = vedo.Plotter(title=title)
        plt.camera.SetPosition((0, 0, artefact_thickness))
        plt.camera.SetFocalPoint((0, -20, 0))
        plt.show(artefact.alpha(0.6), slice_data.lines, plot_points, axes=True)

    sort_order = np.argsort(slice_samples.angle.full_angle)
    slice_samples = slice_samples[sort_order]

    return slice_data, slice_samples


def get_artefact_samples(
    artefact: vedo.Mesh,
    sample_parameters: SampleParameters,
) -> ArtefactInfo:
    """Get a set of samples for the artefact working down the y axis slice by
    slice"""

    tip, analysis_end = get_artefact_y_intersect(artefact=artefact)

    artefact_info = ArtefactInfo(
        end_points=(tip, analysis_end),
    )

    tip_spacing = sample_parameters.tip_spacing.value
    butt_spacing = sample_parameters.butt_spacing.value
    slice_count = sample_parameters.slice_count.value
    point_count = sample_parameters.point_count.value

    # TODO: slicing at the actual bounds doesn't always work reduce the start by
    # one and the end by 1 to make sure

    length_max = tip[1]

    length_min = analysis_end[1]

    overall_length = abs(length_max - length_min)
    portion_to_slice = 100 - tip_spacing - butt_spacing

    # reduce the start by a percentage
    start = length_max - overall_length * tip_spacing / 100

    portion_length = overall_length * portion_to_slice / 100

    inter_slice_distance = portion_length / (slice_count - 1)

    for i in tqdm(range(slice_count), desc="Radial Points Sampling"):
        # work along length from first slice (tip end]) to last slice (butt end)

        slice_location = start - inter_slice_distance * i

        slice_data, slice_samples = get_slice_points(
            artefact=artefact,
            slice_location=slice_location,
            point_count=point_count,
        )

        artefact_info.samples = np.concatenate(
            (artefact_info.samples, slice_samples)
        ).view(np.recarray)

        artefact_info.original_slice_data.append(slice_data)

    return artefact_info



import dataclasses


@dataclasses.dataclass
class SliceCount:
    value: int

    def __post_init__(self):
        self.text: str = f"s{self.value:04d}"


@dataclasses.dataclass
class PointCount:
    value: int

    def __post_init__(self):
        self.text: str = f"p{self.value:04d}"


@dataclasses.dataclass
class TipSpacing:
    value: float

    def __post_init__(self):
        self.text: str = f"ts{self.value:0.3f}"


@dataclasses.dataclass
class ButtSpacing:
    value: float

    def __post_init__(self):
        self.text: str = f"bs{self.value:0.3f}"


@dataclasses.dataclass
class SampleParameters:
    slice_count: SliceCount
    point_count: PointCount
    tip_spacing: TipSpacing
    butt_spacing: ButtSpacing

    def __init__(
        self,
        slice_count: int,
        point_count: int,
        tip_spacing: float,
        butt_spacing: float,
    ):
        self.slice_count = SliceCount(value=slice_count)
        self.point_count = PointCount(value=point_count)
        self.tip_spacing = TipSpacing(value=tip_spacing)
        self.butt_spacing = ButtSpacing(value=butt_spacing)

    def to_csv_dict_data(self):
        return {
            "Slice_Count": self.slice_count.value,
            "Point_Count": self.point_count.value,
            "Tip_Spacing": self.tip_spacing.value,
            "Butt_Spacing": self.butt_spacing.value,
        }



"""
Measurements from Quadrants of original orientated mesh
"""

import dataclasses

from . import Scale
from . import ArtefactSegments


@dataclasses.dataclass
class QuadrantMeasurements:
    """Derive quadrant volume measurements from oriented artefact"""

    artefact_segments: (
        ArtefactSegments  # Attribute storing information about artefact segments
    )

    def __post_init__(self):
        """
        Post-initialization method to calculate additional attributes after the
        object is created.
        """

        # Calculating the original artefact model volume and storing it in a
        # variable.
        self.original_artefact_volume: float = self.artefact_segments.artefact.volume()

        # REFINEMENT: Defining variables for minimum and maximum breadth,
        # length, and thickness values.

        self.breadth_min: float  # Minimum breadth value
        self.breadth_max: float  # Maximum breadth value
        self.length_min: float  # Minimum length value
        self.length_max: float  # Maximum length value
        self.thickness_min: float  # Minimum thickness value
        self.thickness_max: float  # Maximum thickness value

        # Extracting minimum and maximum values for breadth, length, and
        # thickness from artefact bounds.
        (
            self.breadth_min,
            self.breadth_max,
            self.length_min,
            self.length_max,
            self.thickness_min,
            self.thickness_max,
        ) = self.artefact_segments.artefact.bounds()

    def to_volume_data_dict(self):
        return {
            "Artefact_Volume": self.artefact_segments.volume * Scale.VOLUME,
            "Length_Max": self.length_max * Scale.LENGTH,
            "Length_Min": self.length_min * Scale.LENGTH,
            "Thickness_Max": self.thickness_max * Scale.LENGTH,
            "Thickness_Min": self.thickness_min * Scale.LENGTH,
            "Breadth_Max": self.breadth_max * Scale.LENGTH,
            "Breadth_Min": self.breadth_min * Scale.LENGTH,
            "Volume_QI": self.artefact_segments.ventral_right.volume() * Scale.VOLUME,
            "Volume_QII": self.artefact_segments.dorsal_right.volume() * Scale.VOLUME,
            "Volume_QIII": self.artefact_segments.dorsal_left.volume() * Scale.VOLUME,
            "Volume_QIV": self.artefact_segments.ventral_left.volume() * Scale.VOLUME,
        }



from enum import Enum
import dataclasses


class AngularDirection(Enum):  # viewed from artefact base
    clockwise: int = 1
    anticlockwise: int = -1


@dataclasses.dataclass
class FaceAttributes:
    face_centre_angle: float


class QuadrantFace(Enum):
    ventral: FaceAttributes = FaceAttributes(90)
    dorsal: FaceAttributes = FaceAttributes(270)


class Sign(Enum):
    positive: int = 1
    negative: int = -1


@dataclasses.dataclass
class QuadrantAttributes:
    start_angle: float
    end_angle: float
    breadth_sign: Sign
    thickness_sign: Sign
    quadrant_face: QuadrantFace
    angular_direction: AngularDirection


class Quadrant(Enum):
    Q1 = QuadrantAttributes(
        start_angle=0,
        end_angle=QuadrantFace.ventral.value.face_centre_angle,
        breadth_sign=Sign.positive,
        thickness_sign=Sign.positive,
        quadrant_face=QuadrantFace.ventral,
        angular_direction=AngularDirection.clockwise,
    )
    Q2 = QuadrantAttributes(
        start_angle=180,
        end_angle=QuadrantFace.dorsal.value.face_centre_angle,
        breadth_sign=Sign.positive,
        thickness_sign=Sign.negative,
        quadrant_face=QuadrantFace.dorsal,
        angular_direction=AngularDirection.anticlockwise,
    )
    Q3 = QuadrantAttributes(
        start_angle=180,
        end_angle=QuadrantFace.dorsal.value.face_centre_angle,
        breadth_sign=Sign.negative,
        thickness_sign=Sign.negative,
        quadrant_face=QuadrantFace.dorsal,
        angular_direction=AngularDirection.clockwise,
    )
    Q4 = QuadrantAttributes(
        start_angle=360,
        end_angle=QuadrantFace.ventral.value.face_centre_angle,
        breadth_sign=Sign.negative,
        thickness_sign=Sign.positive,
        quadrant_face=QuadrantFace.ventral,
        angular_direction=AngularDirection.anticlockwise,
    )



from ._sample import SampleType
from . import AnalysisMethod

import numpy as np

import dataclasses


@dataclasses.dataclass
class ConstructionData:
    analysis_method: AnalysisMethod
    end_points: tuple[np.ndarray, np.ndarray]

    sample_data: np.ndarray = dataclasses.field(
        default_factory=lambda: np.empty(shape=0, dtype=SampleType).view(np.recarray)
    )



from . import ArtefactInfo
from . import ConstructionData
from . import ArtefactAnalysis
from . import ArtefactSegments

import numpy as np
import vedo

import matplotlib.pyplot as pyplot

from jinja2 import Environment, PackageLoader


def oriented_artefact_plot(
    artefact: vedo.Mesh,
    artefact_name: str,
    orientation_text: str,
):
    title = f"Oriented Artefact Plot: Dimensions for {artefact_name} Oriented {orientation_text}"
    plt = vedo.Plotter(title=title)
    plt.show(artefact, axes=7)

    return artefact.clone()


# The function below defines the Slices Plot. The original Artefact Slice is
# shown as a thick brown line, The Reference Mesh (Endo-/Mean-/Exo-Sym) slice is
# shown as a thick magenta/red/blue line. The X-axis mirrored face of the
# original artefact slice is shown as a thin green line. The mirrored side of
# the original artefact slice is shown as a thin orange line. The mirrored face
# and side of the original artefact slice is shown as a thin purple line.


def slices_plot(
    artefact: vedo.Mesh,
    artefact_info: ArtefactInfo,
    construction_data_list: list[ConstructionData],
):
    (
        breadth_min,
        breadth_max,
        length_min,
        length_max,
        thickness_min,
        thickness_max,
    ) = artefact.bounds()

    breadth_lim = abs(breadth_max - breadth_min) / 2
    thickness_lim = abs(thickness_max - thickness_min) / 2

    ylim_min, ylim_max = (-breadth_lim, breadth_lim)
    xlim_min, xlim_max = (-thickness_lim, thickness_lim)

    samples = artefact_info.samples

    # plot_type = construction_data.analysis_method.text

    slice_length_list = np.unique(samples.point.length)
    slice_length_list[::-1].sort()  # sort in descending order

    slice_count = slice_length_list.size

    fig, axs = pyplot.subplots(1, slice_count, sharey=True, figsize=(50, 7))

    legend_label = []
    fig.autofmt_xdate(rotation=90)
    for slice_number, slice_length in enumerate(slice_length_list):
        point_list_from_artefact_axis = samples[samples.point.length == slice_length]

        axs[slice_number].set_ylim(ylim_min * 1.1, ylim_max * 1.1)
        axs[slice_number].set_xlim(xlim_min * 1.1, xlim_max * 1.1)

        legend_label.append("Artefact Slice")
        axs[slice_number].plot(
            point_list_from_artefact_axis.point.thickness,
            point_list_from_artefact_axis.point.breadth,
            c="brown",
            linewidth=1,
        )

        if True:  # Plot the mirrored X-axis artefact slices
            legend_label.append("Mirror X-Axis")
            axs[slice_number].plot(
                -point_list_from_artefact_axis.point.thickness,
                point_list_from_artefact_axis.point.breadth,
                c="green",
                linewidth=1,
            )

        if True:  # Plot the mirrored Z-axis artefact slices
            legend_label.append("Mirror Z-Axis")
            axs[slice_number].plot(
                point_list_from_artefact_axis.point.thickness,
                -point_list_from_artefact_axis.point.breadth,
                c="orange",
                linewidth=1,
            )

        if True:  # Plot the mirrored X- and Z-axes artefact slices
            legend_label.append("Mirror X- & Z-Axes")
            axs[slice_number].plot(
                -point_list_from_artefact_axis.point.thickness,
                -point_list_from_artefact_axis.point.breadth,
                c="purple",
                linewidth=1,
            )

        for construction_data in construction_data_list:
            construction_samples = construction_data.sample_data

            plot_colour = construction_data.analysis_method.colour

            plot_points = construction_samples[
                construction_samples.point.length == slice_length
            ]

            if True:  # Plot the reference mesh slices
                legend_label.append(construction_data.analysis_method.text)
                axs[slice_number].plot(
                    plot_points.point.thickness,
                    plot_points.point.breadth,
                    c=plot_colour,
                    linewidth=2,
                )

        # TODO: Proper way to set orientation of xticks, but needs work using
        # fig.autofmt_xdate(rotation=90) as an interim
        axs[slice_number].set_xticks(axs[slice_number].get_xticks())
        axs[slice_number].set_xticklabels(
            axs[slice_number].get_xticklabels(), rotation=90, ha="center"
        )

    slice_count = len(artefact_info.original_slice_data)
    title = f"Defining the Geometry of the {construction_data.analysis_method.text} - Radial Distance from Y-Axis (0,0) for {slice_count} Crossectional Slices of {artefact_info.name} from Tip (Left) to Butt (Right)"
    fig.suptitle(title)
    fig.legend(
        list(dict.fromkeys(legend_label)),
        # ["Original", "Endo-Sym", "Exo-Sym", "Mean-Sym"],
        loc="upper center",
        bbox_to_anchor=(0.5, 0.945),
        ncol=7,
        fancybox=True,
        shadow=True,
    )
    fig.text(0.5, 0.04, "Z-Axis = Thickness (mm)", ha="center", va="center")
    fig.text(
        0.02,
        0.5,
        "X-Axis = Breadth (mm)",
        ha="center",
        va="center",
        rotation="vertical",
    )
    pyplot.subplots_adjust(
        left=0.048,
        bottom=0.14,
        right=0.979,
        top=0.864,
        wspace=0.19,
        hspace=None,
    )

    pyplot.show(block=True)


def segment_plot(
    artefact: vedo.Mesh,
    artefact_segments: ArtefactSegments,
):
    artefact_axes = vedo.Axes(artefact)
    title = "Bilateral Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.left.c("yellow").alpha(0.4),
        artefact_segments.right.c("blue").alpha(0.4),
        artefact_axes,
    )
    title = "Bifacial Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.dorsal.c("green").alpha(0.4),
        artefact_segments.ventral.c("red").alpha(0.4),
        artefact_axes,
    )
    title = "QI Ventral-Right Quadrant Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.ventral_right.c("red").alpha(0.4),
        artefact_axes,
    )
    title = "QII Dorsal-Right Quadrant Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.dorsal_right.c("yellow").alpha(0.4),
        artefact_axes,
    )
    title = "QIII Dorsal-Left Quadrant Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.dorsal_left.c("green").alpha(0.4),
        artefact_axes,
    )
    title = "QIV Ventral-Left Quadrant Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.ventral_left.c("blue").alpha(0.4),
        artefact_axes,
    )
    title = "Oriented Quadrants Plot"
    vedo.Plotter(title=title).show(
        artefact.alpha(0.2),
        artefact_segments.dorsal_right.c("red").alpha(0.4),
        artefact_segments.dorsal_left.c("green").alpha(0.4),
        artefact_segments.ventral_right.c("yellow").alpha(0.4),
        artefact_segments.ventral_left.c("blue").alpha(0.4),
        artefact_axes,
    )


def html_output(
    artefact_info: ArtefactInfo,
    artefact_segments: ArtefactSegments,
    construction_data: ConstructionData,
    artefact_analysis: ArtefactAnalysis,
) -> str:
    environment = Environment(loader=PackageLoader("lithikos", "templates/"))
    template = environment.get_template("sample_output.html")
    return template.render(
        artefact_info=artefact_info,
        artefact_segments=artefact_segments,
        construction_data=construction_data,
        artefact_analysis=artefact_analysis,
    )


def txt_output(
    artefact_info: ArtefactInfo,
    artefact_segments: ArtefactSegments,
    construction_data: ConstructionData,
    artefact_analysis: ArtefactAnalysis,
) -> str:
    environment = Environment(loader=PackageLoader("lithikos", "templates/"))
    template = environment.get_template("sample_output.txt")
    return template.render(
        artefact_info=artefact_info,
        artefact_segments=artefact_segments,
        construction_data=construction_data,
        artefact_analysis=artefact_analysis,
    )



import pathlib
import vedo

from . import OrientationMethod
from . import AnalysisMethod
from . import SampleParameters
from . import ArtefactSegments
from . import ConstructionData
from . import ArtefactInfo
from . import QuadrantMeasurements
from . import ArtefactAnalysis
from . import Scale
from . import utils
from . import sampling
from . import analysis

from . import orientation
from . import construction
from . import plotting

from . import logger


def process(
    original_artefact: vedo.Mesh,
    artefact_name: str,
    output_dir: pathlib.Path,
    orientation_methods: list[OrientationMethod],
    analysis_methods: list[AnalysisMethod],
    sample_parameters_list: SampleParameters,
    quadrant_subdir: pathlib.Path,
    artefact_data_file: pathlib.Path,
    reference_data_file: pathlib.Path,
):
    logger.info(
        "Processing artefact: %s",
        artefact_name,
    )
    for orientation_method in orientation_methods:
        orientation_abbr = orientation_method.abbreviation
        orientation_text = orientation_method.text

        logger.info(
            "Processing artefact: %s orientation %s",
            artefact_name,
            orientation_text,
        )

        oriented_artefact_identifier = f"{artefact_name}_{orientation_abbr}"

        oriented_mesh_file_name = f"{oriented_artefact_identifier}_3D-model.obj"
        oriented_mesh_full_file_name = output_dir / oriented_mesh_file_name

        print(oriented_mesh_full_file_name)

        oriented_artefact = orientation.orient_artefact(
            artefact=original_artefact,
            artefact_name=artefact_name,
            orientation_method=orientation_method,
        )

        # If True, this displays the Oriented Artefact Plot with Model Volume
        # and dimensions.

        if False:
            plotting.oriented_artefact_plot(
                artefact=oriented_artefact,
                artefact_name=artefact_name,
                orientation_text=orientation_text,
            )

        (
            breadth_min,
            breadth_max,
            length_min,
            length_max,
            thickness_min,
            thickness_max,
        ) = oriented_artefact.bounds()

        initial_columns = {
            "Artefact_ID": artefact_name,
            "Orientation_Type": orientation_text,
        }

        if False:  # Setting this will calculate the dimension data then end
            continue

        vedo.write(oriented_artefact, oriented_mesh_full_file_name.as_posix())

        if False:
            read_artefact = vedo.Mesh(oriented_mesh_full_file_name.as_posix())
            print(
                f"""original vol: {original_artefact.volume()
                } oriented vol:{oriented_artefact.volume() } read
                vol:{read_artefact.volume()}"""
            )

        oriented_artefact_segments = ArtefactSegments(oriented_artefact)

        quadrant_measurements = QuadrantMeasurements(
            artefact_segments=oriented_artefact_segments,
        )
        artefact_data = (
            initial_columns
            | {
                "Artefact_Surface_Area": oriented_artefact.area() * Scale.AREA,
            }
            | quadrant_measurements.to_volume_data_dict()
        )

        utils.output_csv_data(
            filepath=artefact_data_file,
            data_dict=artefact_data,
        )

        quadrant_output_dir = output_dir / quadrant_subdir

        # TODO: handle error cases
        if not quadrant_output_dir.exists():
            quadrant_output_dir.mkdir(parents=True)
        segment_map = {
            # f'{oriented_artefact_identifier}-face-V.obj':
            #    oriented_artefact_segments.ventral,
            # f'{oriented_artefact_identifier}-face-D.obj':
            #    oriented_artefact_segments.dorsal,
            # f'{oriented_artefact_identifier}-side-R.obj':
            #    oriented_artefact_segments.right,
            # f'{oriented_artefact_identifier}-side-L.obj':
            #    oriented_artefact_segments.left,
            f"{oriented_artefact_identifier}-QI_VR.obj": oriented_artefact_segments.ventral_right,
            f"{oriented_artefact_identifier}-QII_DR.obj": oriented_artefact_segments.dorsal_right,
            f"{oriented_artefact_identifier}-QIII_DL.obj": oriented_artefact_segments.dorsal_left,
            f"{oriented_artefact_identifier}-QIV_VL.obj": oriented_artefact_segments.ventral_left,
        }
        for name, segment in segment_map.items():
            segment_file = quadrant_output_dir / name
            vedo.write(segment, segment_file.as_posix())

        if False:
            # plot segments
            plotting.segment_plot(
                artefact=oriented_artefact,
                artefact_segments=oriented_artefact_segments,
            )

        sample_parameters: SampleParameters
        for sample_parameters in sample_parameters_list:
            artefact_info: ArtefactInfo

            artefact_info = sampling.get_artefact_samples(
                artefact=oriented_artefact,
                sample_parameters=sample_parameters,
            )
            artefact_info.name = artefact_name

            construction_data_list: list[ConstructionData]

            construction_data_list = []
            for analysis_method in analysis_methods:
                construction_data = analysis.analyse_artefact_samples(
                    artefact_info=artefact_info,
                    analysis_method=analysis_method,
                )
                construction_data_list.append(construction_data)

            artefact_info.orientation_method = orientation_method

            slice_count = sample_parameters.slice_count
            point_count = sample_parameters.point_count
            tip_spacing = sample_parameters.tip_spacing
            butt_spacing = sample_parameters.butt_spacing

            analysis_info_text = f"{slice_count.text}-{point_count.text}"
            spacing_info_text = f"{tip_spacing.text}-{butt_spacing.text}"

            if False:
                # This code displays the Slices Plot.
                plotting.slices_plot(
                    artefact=oriented_artefact,
                    artefact_info=artefact_info,
                    construction_data_list=construction_data_list,
                )

            for construction_data in construction_data_list:
                analysis_type = construction_data.analysis_method

                analysis_type_abbr = analysis_type.abbreviation

                analysis_type_text = analysis_type.text

                mesh_colour = analysis_type.colour

                analysis_detail_text = f"{analysis_info_text}_{spacing_info_text}"

                constructed_mesh_file_name = f"{oriented_artefact_identifier}_{analysis_type_abbr}_{analysis_detail_text}.obj"
                html_output_file_name = f"{oriented_artefact_identifier}_{analysis_type_abbr}_{analysis_detail_text}_out.html"
                txt_output_file_name = f"{oriented_artefact_identifier}_{analysis_type_abbr}_{analysis_detail_text}_out.txt"

                sample_constructed_mesh = construction.create_mesh_from_point_data(
                    construction_data
                )

                constructed_result = sample_constructed_mesh

                artefact_analysis = ArtefactAnalysis(
                    artefact_segments=oriented_artefact_segments,
                    constructed_result=constructed_result,
                )
                # If set to True, this code displays the final Reference Mesh
                # Plot before printing the Results to the Terminal.

                if False:
                    title = constructed_mesh_file_name
                    plt = vedo.Plotter(title=title)
                    plt.show(
                        constructed_result.mesh.alpha(0.3).c(mesh_colour),
                    )

                constructed_result.mesh.c(mesh_colour)

                reference_data = (
                    initial_columns
                    | sample_parameters.to_csv_dict_data()
                    | {
                        "Analysis_Type": analysis_type_text,
                    }
                    | {
                        "Reference_Volume": constructed_result.mesh.volume()
                        * Scale.VOLUME,
                    }
                )

                utils.output_csv_data(
                    filepath=reference_data_file,
                    data_dict=reference_data,
                )

                html_output = plotting.html_output(
                    artefact_info=artefact_info,
                    artefact_segments=oriented_artefact_segments,
                    construction_data=construction_data,
                    artefact_analysis=artefact_analysis,
                )

                html_output_full_file_name = output_dir / html_output_file_name

                with open(html_output_full_file_name, "w", encoding="utf-8") as file:
                    file.write(html_output)

                txt_output = plotting.txt_output(
                    artefact_info=artefact_info,
                    artefact_segments=oriented_artefact_segments,
                    construction_data=construction_data,
                    artefact_analysis=artefact_analysis,
                )

                txt_output_full_file_name = output_dir / txt_output_file_name

                with open(txt_output_full_file_name, "w", encoding="utf-8") as file:
                    file.write(txt_output)

                print(txt_output)
                # print(f'Generated mesh has {clean_mesh.npoints} vertices and
                # {clean_mesh.ncells} faces!')
                constructed_mesh_full_file_name = (
                    output_dir / constructed_mesh_file_name
                )
                vedo.write(
                    constructed_result.mesh, constructed_mesh_full_file_name.as_posix()
                )



"""Artefact Utils library supporting analysis of 3D scanned artefacts"""

import logging

from .exceptions import ArtefactUtilsParseException

from .Scale import Scale

from ._sample import (
    # PointXyzType,
    # PointBltType,
    # AngleType,
    SampleType,
)

from .sample_parameters import SampleParameters
from .slice_data import SliceData
from .orientation_method import OrientationMethod
from .analysis_method import AnalysisMethod
from .construction_data import ConstructionData
from .contructed_result import ConstructedResult
from .artefact_info import ArtefactInfo

from .quadrant_data import AngularDirection
from .quadrant_data import QuadrantAttributes
from .quadrant_data import Quadrant

from .measurement_data import MeasurementData

from .artefact_segments import ArtefactSegments
from .quadrant_measurements import QuadrantMeasurements
from .artefact_analysis import ArtefactAnalysis


logger = logging.getLogger(__name__)

__all__ = [
    "AnalysisMethod",
    "AngularDirection",
    "ArtefactAnalysis",
    "ArtefactInfo",
    "ArtefactSegments",
    "ArtefactUtilsParseException",
    "ConstructedResult",
    "ConstructionData",
    "MeasurementData",
    "OrientationMethod",
    "Quadrant",
    "SampleParameters",
    "SampleType",
    "Scale",
    "SliceData",
    "QuadrantAttributes",
    "QuadrantMeasurements",
]
